import sys
from inspect import getfullargspec, signature, Signature, Parameter
from typing import Sequence, Callable, Optional, get_type_hints

from .core import Arg, NotSet, DEPS_METHOD, FunctionArgs, NotSetType, INJECTION_ARG, \
    QUERY_WRAPPED_METHOD, InjectionClauses, INJECTED, init_arg_injection, signature_injection


def is_builtin_cls(annotation) -> [NotSetType, bool]:
    return NotSet if annotation is NotSet else annotation.__module__ == 'builtins'


def get_deps_from_signature(signature: Signature, exclude_first=False) -> Sequence[Arg]:
    result = []

    first = True
    for name, param in signature.parameters.items():
        if first and exclude_first:
            exclude_first = False
            continue
        if param.kind in [Parameter.VAR_POSITIONAL, Parameter.VAR_KEYWORD]:
            continue
        annotation = NotSet if param.annotation is Parameter.empty else param.annotation
        default = NotSet if param.default is Parameter.empty else param.default
        result.append(Arg(name, annotation, default))
    return result



def get_func_result(func: Callable) -> Sequence[Arg]:
    assert isinstance(func, Callable), \
        '"{}" must be Callable'.format(func.__name__)

    sign = signature(func)
    return NotSet if sign.return_annotation is Parameter.empty else sign.return_annotation


def get_func_args(func: Callable) -> Sequence[Arg]:
    assert isinstance(func, Callable), \
        '"{}" must be Callable'.format(func.__name__)

    return get_deps_from_signature(signature(func))


def get_method_args(cls, method_name) -> Sequence[Arg]:
    method = getattr(cls, method_name)

    assert method, 'Missing method {1} on class {0}'.format(cls.__name__, method_name)

    assert isinstance(method, Callable), \
        '"{0}.{1}" must be Callable'.format(cls.__name__, method_name)

    return get_deps_from_signature(signature(method), True)


def has_init_arg_injection(cls) -> bool:
    if not init_arg_injection:
        return False

    init_method = getattr(cls, '__init__')

    if not init_method:
        return False

    spec = getfullargspec(init_method)

    try:
        if spec.kwonlyargs and spec.kwonlyargs.index(INJECTION_ARG) >= 0:
            return True
    except ValueError:
        pass

    return False


def get_signature_deps(cls) -> Sequence[Arg]:
    args = []

    if not signature_injection:
        return args

    # Base deps
    for base_cls in cls.__bases__:
        if base_cls.__module__ != 'builtins':
            args += get_signature_deps(base_cls)

    # Current deps
    method = getattr(cls, DEPS_METHOD, None)
    if method:
        args += get_method_args(cls, DEPS_METHOD)

    return args


def get_init_deps(cls) -> Sequence[Arg]:
    init_method = getattr(cls, '__init__')
    if not init_method:
        return ()

    ## Access real __init__ when using injection decorator
    if has_init_arg_injection(cls):
        real_init = init_method(QUERY_WRAPPED_METHOD)
        return get_deps_from_signature(signature(real_init), True)

    else:
        return get_method_args(cls, '__init__')


def get_attribute_deps(cls) -> Sequence[Arg]:

    # Field annotations available from Python 3.6
    if (sys.version_info.major == 3 and sys.version_info.minor < 6):
        return []

    result = []

    hints = get_type_hints(cls)
    for att in (dir(cls)):
        if att[:2] != '__' and getattr(cls, att, None) == INJECTED:
            result.append(Arg(att, hints.get(att, NotSet), NotSet))

    return result

def get_class_dependencies(cls) -> Sequence[Arg]:

    # __init__ args must come first to know the position of its arguments
    args = get_init_deps(cls)

    args += get_attribute_deps(cls)

    if signature_injection:
        args += get_signature_deps(cls)

    return args



def subclass_inject(cls, args, kwargs, kwinject):

    #Injector = type(cls.__name__, (cls,), {'__metaclass__': WirinjMeta})
    class Injector(cls):
        def __new__(injector_cls, *args, **kwargs):

            super_new = super(Injector, injector_cls).__new__
            if super_new is object.__new__:
                instance = super_new(cls)
            else:
                instance = super_new(cls, *args, **kwargs)

            #Inject
            for name, value in kwinject.items():
                setattr(instance, name, value)

            # Init must be called ourself since we are creating a cls instance and not a injector_cls instance.
            instance.__init__(*args, **kwargs)

            return instance
    return Injector(*args, **kwargs)


def instantiate_class(cls, instance_args: Optional[FunctionArgs] = None, **deps):

    priv_args = get_attribute_deps(cls)

    if signature_injection:
        priv_args += get_signature_deps(cls)

    # Collect priv args
    priv_deps = {}
    for priv_arg in priv_args:
        key = priv_arg.name
        if key in deps:
            priv_deps[key] = deps[key]

    # Collect public args
    pub_deps = {}
    for key, value in deps.items():
        if key not in priv_deps:
            pub_deps[key] = value

    # Instance args and kwargs
    if instance_args:
        args = instance_args.args
        kwargs = instance_args.kwargs
    else:
        args = ()
        kwargs = {}

    # Instance
    if priv_deps:
        if has_init_arg_injection(cls):
            return cls(*args, **{INJECTION_ARG: priv_deps, **pub_deps, **kwargs})
        else:
            return subclass_inject(cls, args, {**pub_deps, **kwargs}, priv_deps)
    else:
        return cls(*args, **{**pub_deps, **kwargs})



from .core import logger, Arg, Dependency, Locator, INJECTED
from .autowiring import AutowiringReport, Autowiring
from .decorators import deps, inject
from .definition import Definitions, DependencyBuilder, Singleton, Factory, Instance, CustomSingleton, CustomInstance,\
    CustomFactory
from .injector import Injector
from .locators import Locator, LocatorCache, LocatorChain

__all__ = [x for x in dir() if not x.startswith('_')]



from logging import ERROR, INFO, DEBUG
from typing import Union, Sequence, Callable, Optional, Dict, Tuple

from .core import logger, Arg, Dependency, NotSet, Locator, SEPARATOR_OPEN, SEPARATOR_CLOSE, FunctionArgs, \
    filter_direct_args, InjectionClauses
from .errors import MissingDependenciesError
from .introspect import get_func_args
from .locators import LocatorChain, LocatorCache


class NotFoundType(type):
    def __str__(self):
        return 'NotFound'


class NotFound(metaclass=NotFoundType):
    pass


class FailedType(type):
    def __str__(self):
        return 'Failed'


class Failed(metaclass=FailedType):
    pass


class CreationNode:
    def __init__(self,
                 arg: Arg,
                 dep: Union[Dependency, NotFoundType, None],
                 childs: Sequence['CreationNode'] = (),
                 ):
        self.arg = arg
        self.dep = dep
        self.childs = childs
        self.instance = None

    def get_params(self):
        result = {}
        for child in self.childs:
            result[child.arg.name] = child.instance
        return result

    def __str__(self) -> str:
        return "{}{}".format(
            str(self.arg),
            ' *** {} ***'.format(str(self.dep)) if self.dep is NotFound else '',
        )


class CreationNodeReference:
    def __init__(self, creation_path: Sequence[CreationNode], node: CreationNode):
        self.creation_path = creation_path
        self.node = node

    def __str__(self) -> str:
        path = ''
        for entry in self.creation_path:
            path += '    '

        return "{}{}".format(
            path,
            str(self.node),
        )


def get_tree_index(root: CreationNode, creation_path=()) -> Sequence[CreationNodeReference]:
    result = []
    for child in root.childs:
        result += get_tree_index(child, (list(creation_path) + [root.arg]))

    result += [CreationNodeReference(creation_path, root)]

    return result


def get_creation_args(node_childs: Sequence[CreationNode]) -> Dict:
    result = {}
    for child in node_childs:
        result[child.arg.name] = child.instance
    return result


def log_creation_tree(tree: CreationNode, log_level=INFO, msg=None):
    if not logger.isEnabledFor(log_level):
        return

    dep_flat = get_tree_index(tree)

    logger.log(log_level, SEPARATOR_OPEN)
    if msg:
        logger.log(log_level, msg)
    for node_ref in dep_flat:
        logger.log(log_level, str(node_ref))
    logger.log(log_level, SEPARATOR_CLOSE)


def _after_tree_creation(success: bool, tree: CreationNode):
    if success:
        log_creation_tree(tree, DEBUG)
    else:
        log_creation_tree(tree, ERROR, 'Missing dependencies:')
        raise MissingDependenciesError('Missing dependencies.')


def creation_path_as_text(parent_path: Sequence[Arg], creation_node: CreationNode, suffix=''):
    path = ''
    for entry in parent_path:
        path += str(entry) + ' -> '

    path += str(creation_node.arg)

    if suffix:
        path += ' [{}]'.format(suffix)

    return path


def has_a_valid_default(arg: Arg):
    if arg.default is NotSet:
        return False

    if arg.cls != NotSet and arg.default in InjectionClauses:
        return False

    return True


class DefaultDependency(Dependency):

    def __init__(self, default):
        self.default = default

    def get_instance(self, instance_args=None, **deps):
        return self.default


class Injector:
    """
    Dependency injection service.
    """

    def __init__(self, *dependencies: Locator, cached=True):
        """
        @param dependencies: one or more Locator objects such as Dependencies or Autowiring which will be queried
        by the injector object to locate dependencies. If two Locators contain the same dependency, the first takes
        precedence.
        @param cached: if True, cached copies of already located dependecy managers (Dependency) are kept to save time.
        """

        assert dependencies, '{0} requires at least one {1}'.format(Injector.__name__, Locator.__name__)

        if len(dependencies) == 1:
            deps = dependencies[0]
        else:
            deps = LocatorChain(*dependencies)

        self.locator = LocatorCache(deps) if cached else deps

        self.locator.initialize(self)

    def get(self, cls, *args, **kwargs):
        success, root = self._create_node((), Arg(None, cls), FunctionArgs(args, kwargs))
        _after_tree_creation(success, root)
        return root.instance

    def call(self, func: Callable, *args, **kwargs):
        injected_args = self._get_function_args(func, args, kwargs)
        return func(*args, **{**injected_args, **kwargs})

    def get_wrapper(self, func: Callable):
        def func_wrapper(*args, **kwargs):
            injected_args = self._get_function_args(func, args, kwargs)
            return func(*args, **{**injected_args, **kwargs})

        return func_wrapper

    def _get_function_args(self, func: Callable, args, kwargs):
        fn_args = get_func_args(func)
        injectable_args = filter_direct_args(fn_args, args, kwargs)
        return self._create_virtual_node(injectable_args, Arg(func.__name__, type(func)))

    def _create_childs(self, arg_list: Optional[Sequence[Arg]], creation_path: Sequence[Arg]) -> Tuple[
        bool, Sequence[CreationNode]]:
        """
        @return: Returns two values. The first value is True if the instantiation of all childs succeded. The second
        is the CreationNode list.
        """
        if not arg_list:
            return True, ()

        childs = []
        success = True
        for arg in arg_list:
            child_success, child = self._create_node(creation_path, arg)
            if not child_success:
                success = False
            childs.append(child)
        return success, childs

    def _create_node(self, parent_path: Sequence[Arg], arg: Arg, instance_args: Optional[FunctionArgs] = None) -> Tuple[
        bool, CreationNode]:
        """
        @return: Returns two values. The first value is True if the instantiation succeded. The second is the new
        node. When the Locator cannot find a dependency, the field 'dep' will be set as NotFound.
        """
        current_path = tuple(parent_path) + (arg,)

        # Find in the locator
        dep = self.locator.get(current_path)
        if not dep:
            if has_a_valid_default(arg):
                dep = DefaultDependency(arg.default)
            else:
                return False, CreationNode(current_path[-1], NotFound)

        # Get dependency args
        dep_args = dep.get_dependencies()

        # With args
        if dep_args:
            # Update path with actual class
            cls = dep.get_class()
            if cls is not NotSet and cls != arg.cls:
                current_path = list(parent_path) + [Arg(arg.name, cls, arg.default)]

            # Remove instance args
            if instance_args:
                dep_args = filter_direct_args(dep_args, instance_args.args, instance_args.kwargs)

            # Create childs
            childs_success, childs = self._create_childs(dep_args, current_path)

            # Creation params
            params = get_creation_args(childs)

        # With no args
        else:
            childs = ()
            params = {}
            childs_success = True

        # New CreationNode
        creation_node = CreationNode(current_path[-1], dep, childs)

        # Missing childs dependencies. Return Node with no instance.
        if not childs_success:
            return False, creation_node

        # All dependencies fulfilled. Create instance
        try:
            creation_node.instance = dep.get_instance(instance_args, **params)
        except BaseException as ex:
            creation_node.instance = Failed
            logger.fatal(SEPARATOR_OPEN)
            logger.fatal('Instantiation ERROR:')
            logger.fatal(creation_path_as_text(parent_path, creation_node, ex.__class__.__name__))
            logger.fatal(SEPARATOR_CLOSE)
            raise ex

        # Return full node
        return True, creation_node

    def _create_virtual_node(self, args: Sequence[Arg], root_arg: Arg):

        # Create tree of dependencies
        success, childs = self._create_childs(args, [root_arg])
        root = CreationNode(root_arg, None, childs)
        _after_tree_creation(success, root)
        return root.get_params()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("VERSION", "r") as fh:
    version = fh.read().strip()

setuptools.setup(
    name="wirinj",
    version=version,
    author="David Cimadevilla",
    author_email="dev.davidcim@gmail.com",
    description="A comfy dependency injection library.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/davidcim/wirinj",
    packages=setuptools.find_packages(exclude=['tests*', 'examples*', 'build*']),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



from abc import abstractmethod
from typing import List, Callable, Optional, Dict

from .core import Arg, NotSet, Locator, Dependency
from .dependencies import FactoryDependency, InstanceDependency, SingletonWrapper, ValueDependency, \
    CustomInstanceDependency
from .injector import Injector
from .tools import is_typing_type, get_typing_args


class DefinitionFinder:
    def __init__(self, definitions: Dict):
        assert isinstance(definitions, dict)
        self.definitions = definitions

    def __call__(self, creation_path: List[Arg]):

        result_key = None
        result_value = None
        def_len = 0

        for key, value in self.definitions.items():
            definition = key if isinstance(key, list) or isinstance(key, tuple) else (key,)

            if len(definition) > len(creation_path):
                continue

            if len(definition) <= def_len:
                continue

            match = True
            for i, def_item in enumerate(reversed(definition)):
                path_item = creation_path[-i - 1]
                if not (isinstance(def_item, str) and def_item == path_item.name or def_item == path_item.cls):
                    match = False
                    break
            if match:
                result_key = key
                result_value = value
                def_len = len(definition)

        return (NotSet, NotSet) if def_len == 0 else (result_key, result_value)


class DependencyBuilder:

    @abstractmethod
    def create(self, creation_path: List[Arg], injector: Injector):
        pass


class CustomInstance(DependencyBuilder):
    def __init__(self, creator: Callable, cls=None):
        self.creator = creator
        self.cls = cls

    def create(self, creation_path: List[Arg], injector: Injector):
        if not self.cls and creation_path and creation_path[-1].cls is not NotSet:
            cls = creation_path[-1].cls
        else:
            cls = self.cls
        return CustomInstanceDependency(self.creator, cls)


class CustomSingleton(DependencyBuilder):
    def __init__(self, creator: Callable, cls=None):
        self.creator = creator
        self.cls = cls

    def create(self, creation_path: List[Arg], injector: Injector):
        return SingletonWrapper(CustomInstanceDependency(self.creator, self.cls))


CustomFactory = CustomSingleton


class Factory(DependencyBuilder):
    def __init__(self, cls=None):
        self.cls = cls

    def create(self, creation_path: List[Arg], injector: Injector):
        if self.cls is None:
            last = creation_path[-1]
            assert is_typing_type(last.cls), \
                'Factory without params needs Type[YourClass] as last element in the definition path'
            cls = get_typing_args(last.cls)
        else:
            cls = self.cls

        return SingletonWrapper(FactoryDependency(cls, injector))


class Instance(DependencyBuilder):
    def __init__(self, cls=None):
        self.cls = cls

    def create(self, creation_path: List[Arg], injector: Injector):
        if self.cls is None:
            last = creation_path[-1]
            assert isinstance(last.cls, type), \
                'Instance without params needs YourClass as last element in the definition path'
            cls = last.cls
        else:
            cls = self.cls

        return InstanceDependency(cls)


class Singleton(DependencyBuilder):
    def __init__(self, cls=None):
        self.cls = cls

    def create(self, creation_path: List[Arg], injector: Injector):
        if self.cls is None:
            last = creation_path[-1]
            assert isinstance(last.cls, type), \
                'Singleton without params needs YourClass as last element in the definition path'
            cls = last.cls
        else:
            cls = self.cls

        return SingletonWrapper(InstanceDependency(cls))


class Definitions(Locator):
    def __init__(self, *definitions: Dict):

        assert len(definitions), 'You have to provide at least one definition dict'

        defs = {}
        for df in definitions:
            defs = {**defs, **df}

        self.finder = DefinitionFinder(defs)
        self.injector = None  # type: Optional[Injector]

    def initialize(self, injector):
        self.injector = injector

    def get(self, creation_path: List[Arg]) -> Optional[Dependency]:
        assert self.injector is not None

        _, value = self.finder(creation_path)
        if value is NotSet:
            return None
        if isinstance(value, Dependency):
            return value
        if isinstance(value, DependencyBuilder):
            return value.create(creation_path, self.injector)
        else:
            return ValueDependency(value)
